#include <parallel.h>

#if 0
#define DEBUG_transmit_matrix_host
#define DEBUG_receive_matrix_host
#define DEBUG_transmit_col_vector_host
#define DEBUG_receive_col_vector_node
#define DEBUG_transmit_row_vector_node
#define DEBUG_receive_row_vector_host
#endif

void transmit_intro_letter_host(ZIP_MAILER *Mailer,
                                long Row_Dis_Type, long Col_Dis_Type,
                                long Row_Extra_param1, long Row_Extra_param2,
                                long Col_Extra_param1, long Col_Extra_param2,
                                long Strategy)
{
#define NLONGS 11

   char *Letter;
   long *Header;

   int Letter_Size = NLONGS * sizeof(long);

#ifdef DEBUG_TRANSMIT_INTRO
   print("transmit_intro_letter(): entry...");
#endif

   Zip_push(Mailer);

   Letter    = Zip_malloc(Letter_Size);
   Header    = (long *) Letter;
   Header[0] = INTRO_LETTER;
   Header[1] = (long) Letter_Size;
   Header[2] = NLONGS;
   Header[3] = 0L;

   Header[4] = (long)Row_Dis_Type;
   Header[5] = (long)Col_Dis_Type;

   Header[6] = Row_Extra_param1;
   Header[7] = Row_Extra_param2;
   Header[8] = Col_Extra_param1;
   Header[9] = Col_Extra_param2;

   Header[10] = Strategy;

   ctohl(Header, NLONGS);

   G2_send(Letter, 0, 0);

   Zip_pop();
#undef NLONGS
}

void receive_intro_letter_node(ZIP_MAILER *Mailer,
                               short *Row_Dis_Type, short *Col_Dis_Type,
                               long *Row_Extra_param1, long *Row_Extra_param2,
                               long *Col_Extra_param1, long *Col_Extra_param2,
                               long *Strategy)
{
   char *Subroutine_name = "receive_intro_letter_node():";
   int   P;
   int   Q;
   int   p;
   int   q;
   int   PQ;

   char *Letter;
   int   Letter_Size;

   long *Header;
   long *Dis_Header;
   long *Mat_Data_Header;

   int   G2_pm;
   int   serror;

   Zip_push(Mailer);

   G2_pq(p, q);
   G2_PQ(P, Q);

   PQ = P * Q;

   G2_pm = (!p) && (!q);

   if (G2_pm)
   {
#ifdef DEBUG_RECEIVE_INTRO
      print("I am G2 PM; waiting for letter from HOST");
#endif
      Letter      = (char *) G2_recvb(-1, -1);
      Letter_Size = ((long *)Letter)[1];
   }
   else
   {
      Letter_Size = 0;
      Letter      = NULL;
   }

   /* fanout: */

   if (PQ > 1)
      serror = G2_fanout(&Letter, &Letter_Size, 0, 0);

   Header = (long *) Letter;

#ifdef DEBUG_RECEIVE_INTRO
   print("receive_intro_letter(): After fanout\n");
#endif

   if (*Header != INTRO_LETTER)
   {
      print("csparsetest1a: anticipated INTRO_LETTER=%ld, received %ld",
            (long)INTRO_LETTER, *Header);

      exit(1);
   }

   Dis_Header   = Header + 4;
   *Row_Dis_Type = (short) Dis_Header[0];
   *Col_Dis_Type = (short) Dis_Header[1];

   Mat_Data_Header = (Dis_Header + 2);

   /* data for one and two-parameter distributions */
   *Row_Extra_param1 = Mat_Data_Header[0]; /* "B" */
   *Row_Extra_param2 = Mat_Data_Header[1]; /* "S" */
   *Col_Extra_param1 = Mat_Data_Header[2]; /* "B" */
   *Col_Extra_param2 = Mat_Data_Header[3]; /* "S" */

   *Strategy = Mat_Data_Header[4];

   if (G2_pm)
      Zip_free(Letter);
   else
      free(Letter);

   Zip_pop();
}

transmit_command_host(ZIP_MAILER *Mailer, int Command)
{
#define NLONGS 5

   char *Letter;
   long *Header;

   int Letter_Size = NLONGS * sizeof(long);

#ifdef DEBUG_TRANSMIT_COMMAND
   print("transmit_command_letter(): entry...");
#endif

   Zip_push(Mailer);

   Letter    = Zip_malloc(Letter_Size);
   Header    = (long *) Letter;
   Header[0] = COMMAND_LETTER;
   Header[1] = (long) Letter_Size;
   Header[2] = NLONGS;
   Header[3] = 0L;

   Header[4] = (long) Command;

   ctohl(Header, NLONGS);

   G2_send(Letter, 0, 0);

   Zip_pop();
#undef NLONGS
}

int receive_command_node(ZIP_MAILER *Mailer)
{
   int   P;
   int   Q;
   int   p;
   int   q;
   int   PQ;

   char *Letter;
   int   Letter_Size;

   long *Header;
   long *Dis_Header;
   long *Mat_Data_Header;

   int   G2_pm;
   int   serror;
   int   Command;

   Zip_push(Mailer);

   G2_pq(p, q);
   G2_PQ(P, Q);

   PQ = P * Q;
   G2_pm = (!p) && (!q);

   if (G2_pm)
   {
#ifdef DEBUG_RECEIVE_COMMAND
      print("I am G2 PM; waiting for letter from HOST");
#endif
      Letter      = (char *) G2_recvb(-1, -1);
      Letter_Size = ((long *)Letter)[1];
   }
   else
   {
      Letter_Size = 0;
      Letter      = NULL;
   }

   /* fanout: */
   if (PQ > 1)
      serror = G2_fanout(&Letter, &Letter_Size, 0, 0);

   Header = (long *) Letter;

#ifdef DEBUG_RECEIVE_COMMAND
   print("receive_command_letter(): After fanout\n");
#endif

   if (*Header == COMMAND_LETTER)
   {
      Command = (int) Header[4];
   }
   else
   {
      print("csparsetest1a: anticipated COMMAND_LETTER=%ld, received %ld",
            (long)COMMAND_LETTER, *Header);

      exit(1);
   }

   if (G2_pm)
      Zip_free(Letter);
   else
      free(Letter);

   Zip_pop();

   return(Command);
}

void transmit_matrix_host(ZIP_MAILER *Mailer, matrix_host_bin *Matrix)
{
#define NLONGS 8

   char      *Subroutine_name = "transmit_matrix_host():";
   char      *Letter;
   int        Letter_Size;

   long      *Header;

   long      *Matrix_i;
   long      *Matrix_j;

   double    *Matrix_Value;

   int        M = Matrix -> M;
   int        N = Matrix -> N;

   int        P = Matrix -> P;
   int        Q = Matrix -> Q;

   int        p;
   int        q;

   int      **A_element_Count = Matrix -> A_element_Count;
   long       Matrix_element_Count;
   int        Index;
   element ***A = Matrix -> A;
   element   *Pointer;

#ifdef DEBUG_transmit_matrix_host
   print("%s entry...", Subroutine_name);
#endif

   Zip_push(Mailer);

   for (p = 0; p < P; p++)
      for (q = 0; q < Q; q++)
      {
         Matrix_element_Count = A_element_Count[p][q];

         Letter_Size = NLONGS * sizeof(long)
                       + Matrix_element_Count
                         * ((2 * sizeof(long)) + sizeof(double));

         Letter    = Zip_malloc(Letter_Size);
         Header    = (long *) Letter;
         Header[0] = MATRIX_LETTER;
         Header[1] = (long)Letter_Size;
         Header[2] = NLONGS;
         Header[3] = 0L;

         Header[4] = M;
         Header[5] = N;

         Header[6] = Matrix_element_Count;
         Header[7] = 0L;

         htocl(Header, NLONGS);

#ifdef DEBUG_transmit_matrix_host
         print("%s Matrix_element_Count = %d\n", Subroutine_name,
               Matrix_element_Count);
#endif
         if (Matrix_element_Count > 0)
         {
            Matrix_i = &(Header[NLONGS]);
            Matrix_j = &(Matrix_i[Matrix_element_Count]);
            Matrix_Value = (double *) &(Matrix_j[Matrix_element_Count]);

            Index = 0;
            Pointer = A[p][q];
            do
            {
#ifdef DEBUG_transmit_matrix_host
               print("%s Pointer = 0x%lx\n", Subroutine_name, Pointer);
               print("%s (%d, %d, %d, %d) = %lf\n", Subroutine_name, p, q,
                     Pointer -> i, Pointer -> j, Pointer -> Real);
#endif
               Matrix_i[Index] = Pointer -> i;
               Matrix_j[Index] = Pointer -> j;
               Matrix_Value[Index] = Pointer -> Real;
               Index++;
               Pointer = Pointer -> Next;
            } while (Pointer != NULL);

            if (Index < Matrix_element_Count)
            {
               print("%s Lost matrix entries\n", Subroutine_name);
               print("Index = %d, Matrix_element_Count = %d\n", Index,
                     Matrix_element_Count);
            }
            htocl(Matrix_i, Matrix_element_Count);
            htocl(Matrix_j, Matrix_element_Count);
            htocd(Matrix_Value, Matrix_element_Count);
         }

         G2_send(Letter, p, q);
      }

   Zip_pop();

#undef NLONGS
}

void receive_matrix_host(ZIP_MAILER *Mailer, matrix_host_bin *Matrix)
{
#define NLONGS 8

   char      *Subroutine_name = "receive_matrix_host():";
   char      *Letter;

   long      *Header;

   long      *Matrix_i;
   long      *Matrix_j;

   double    *Matrix_Value;

   int        M = Matrix -> M;
   int        N = Matrix -> N;

   int        P = Matrix -> P;
   int        Q = Matrix -> Q;

   int        p;
   int        q;

   int        I;
   int        I_offset;
   int        J;
   int        J_offset;
   int        Offset;

   int        i;

   int      (*Row_Inverse_Distribution)();
   int      (*Col_Inverse_Distribution)();
   void      *Row_Extra = Matrix -> Row_Extra;
   void      *Col_Extra = Matrix -> Col_Extra;

   int       *Row_major_LU_element_Count = Matrix -> Row_major_LU_element_Count;
   element  **Row_major_LU = Matrix -> Row_major_LU;

   long       Matrix_element_Count;

   element   *Pointer;
   element   *Pool = Matrix -> Pool;

#ifdef DEBUG_receive_matrix_host
   print("%s entry...", Subroutine_name);
#endif

#if 0
   /* four quater partition */
   /* A -B */
   /* B A */
/*
 * I = Ihat;
 * J = Jhat;
 */
   Offset = M >> 1;
#else
   /* Each complex value is maped into a 2x2 real matrix */
   /* This retains the ordering */
/*
 * I = Ihat << 1;
 * J = Jhat << 1;
 */
   Offset = 1;
#endif

   Row_Inverse_Distribution = Matrix -> Row_Inverse_Distribution;
   Col_Inverse_Distribution = Matrix -> Col_Inverse_Distribution;

   Zip_push(Mailer);

   for (p = 0; p < P; p++)
      for (q = 0; q < Q; q++)
      {
         Letter = G2_recvb(p, q);

         Header = (long *) Letter;
         if (*Header != MATRIX_LETTER)
         {
            print("%s anticipated MATRIX_LETTER=%ld, received %ld",
                  Subroutine_name, (long)MATRIX_LETTER, *Header);
            exit(1);
         }

         Matrix_element_Count = Header[6] + Header[7];
         print("Matrix_element_Count = %d\n", Matrix_element_Count);

         Matrix_i = &(Header[NLONGS]);
         Matrix_j = &(Matrix_i[Matrix_element_Count]);
         Matrix_Value = (double *) &(Matrix_j[Matrix_element_Count]);

         for (i = 0; i < Matrix_element_Count; i++)
         {
#if 0
            I = (*Row_Inverse_Distribution)(p, (index) Matrix_i[i], P, M,
                                            Row_Extra);
            J = (*Col_Inverse_Distribution)(q, (index) Matrix_j[i], Q, N,
                                            Col_Extra);
#else
            I = Matrix_i[i];
            J = Matrix_j[i];
#endif
            I_offset = I & 0x00000001;
            I >>= 1;
            J_offset = J & 0x00000001;
            J >>= 1;

#ifdef DEBUG_receive_matrix_host
            print("%s (%d, %d, %d, %d) (%d, %d) = %g\n", Subroutine_name, p, q,
                  Matrix_i[i], Matrix_j[i], I, J, Matrix_Value[i]);
#endif
            if ((I_offset == 0) && (J_offset == 0))
            {
#ifdef DEBUG_receive_matrix_host
               print("%s before add_element_real()\n", Subroutine_name);
#endif
               add_element_real(&Pool, &(Row_major_LU[I]),
                                &(Row_major_LU_element_Count[I]),
                                I, J, Matrix_Value[i]);
            }
            else if ((I_offset == 0) && (J_offset == 1))
            {
#ifdef DEBUG_receive_matrix_host
               print("%s before add_element_imag()\n", Subroutine_name);
#endif
               add_element_imag(&Pool, &(Row_major_LU[I]),
                                &(Row_major_LU_element_Count[I]),
                                I, J, Matrix_Value[i]);
            }
            else if ((I_offset == 1) && (J_offset == 0))
            {
#ifdef DEBUG_receive_matrix_host
               print("%s before add_element_imag()\n", Subroutine_name);
#endif
               add_element_imag(&Pool, &(Row_major_LU[I]),
                                &(Row_major_LU_element_Count[I]),
                                I, J, Matrix_Value[i]);
            }
            else if ((I_offset == 1) && (J_offset == 1))
            {
#ifdef DEBUG_receive_matrix_host
               print("%s before add_element_real()\n", Subroutine_name);
#endif
               add_element_real(&Pool, &(Row_major_LU[I]),
                                &(Row_major_LU_element_Count[I]),
                                I, J, Matrix_Value[i]);
            }
#ifdef DEBUG_receive_matrix_host
            print("%s Row_major_LU_element_Count[%d] = %d\n", Subroutine_name,
                  I, Row_major_LU_element_Count[I]);
#endif
         }
         Zip_free(Letter);
      }

   Zip_pop();

   Matrix -> Pool = Pool;
#undef NLONGS
}

void receive_matrix_node(ZIP_MAILER *Mailer, matrix_node_bin *Matrix)
{
   char   *Subroutine_name = "receive_unfactored_matrix_letter():";
   int     p;
   int     q;

   char   *Letter;

   long   *Header;
   long   *Mat_Data_Header;
   long   *Matrix_i;
   long   *Matrix_j;
   double *Matrix_Value;

   long    Matrix_element_Count;

   Zip_push(Mailer);

   G2_pq(p, q);

   Letter      = (char *) G2_recvb(-1, -1);

   Header      = (long *)Letter;

   if (*Header != MATRIX_LETTER)
   {
      print("csparsetest1a: anticipated MATRIX_LETTER=%ld, received %ld",
            (long)MATRIX_LETTER, *Header);

      exit(1);
   }

   Mat_Data_Header = Header + 4;
   Matrix_i = Mat_Data_Header + 4;
   Matrix_j = Matrix_i + Mat_Data_Header[2];
   Matrix_Value = (double *) (Matrix_j + Mat_Data_Header[2]);

   Matrix -> M = Mat_Data_Header[0];
   Matrix -> N = Mat_Data_Header[1];

   if (Matrix -> Array_len < Mat_Data_Header[2])
   {
      if (Matrix -> Array_len > 0)
      {
         free(Matrix -> i);
         free(Matrix -> j);
         free(Matrix -> Value);
      }

      Matrix -> Array_len = Mat_Data_Header[2];
      Matrix -> i = (long *) malloc(Mat_Data_Header[2] * sizeof(long));
      Matrix -> j = (long *) malloc(Mat_Data_Header[2] * sizeof(long));
      Matrix -> Value = (double *) malloc(Mat_Data_Header[2] * sizeof(double));

      if (Matrix -> Value == NULL)
      {
         print("%s Out of memory!\n", Subroutine_name);
         exit(1);
      }
   }

   Matrix -> num_element = Mat_Data_Header[2];

   bcopy(Matrix_i, Matrix -> i, Mat_Data_Header[2] * sizeof(long));
   bcopy(Matrix_j, Matrix -> j, Mat_Data_Header[2] * sizeof(long));
   bcopy(Matrix_Value, Matrix -> Value, Mat_Data_Header[2] * sizeof(double));

   Zip_free(Letter);

   Zip_pop();
}

void transmit_matrix_node(ZIP_MAILER *Mailer, matrix_node_bin *Matrix)
{
#define NLONGS   8

   char   *Subroutine_name = "transmit_matrix_node():";

   int     Letter_Size;
   char   *Letter;

   long   *Header;
   long   *Matrix_i;
   long   *Matrix_j;
   double *Matrix_Value;

   long    Matrix_element_Count = Matrix -> num_element;

   Zip_push(Mailer);

   Letter_Size = NLONGS * sizeof(long)
                 + Matrix_element_Count * ((2 * sizeof(long)) + sizeof(double));

   Letter    = Zip_malloc(Letter_Size);

   Header    = (long *)Letter;

   Header[0] = MATRIX_LETTER;
   Header[1] = (long)Letter_Size;
   Header[2] = NLONGS;
   Header[3] = 0L;

   Header[4] = Matrix -> m;
   Header[5] = Matrix -> n;
   Header[6] = Matrix -> U_elements_count;
   Header[7] = Matrix -> L_elements_count;

   ctohl(Letter, NLONGS);

   Matrix_i = &(Header[NLONGS]);
   Matrix_j = &(Matrix_i[Matrix_element_Count]);
   Matrix_Value = (double *) &(Matrix_j[Matrix_element_Count]);

   bcopy(Matrix -> i, Matrix_i, Matrix_element_Count * sizeof(long));
   ctohl(Matrix_i, Matrix_element_Count);

   bcopy(Matrix -> j, Matrix_j, Matrix_element_Count * sizeof(long));
   ctohl(Matrix_j, Matrix_element_Count);

   bcopy(Matrix -> Value, Matrix_Value, Matrix_element_Count * sizeof(double));
   ctohd(Matrix_Value, Matrix_element_Count);

   G2_send(Letter, -1, -1);

   Zip_pop();

#undef NLONGS
}

void transmit_timing_node(ZIP_MAILER *Mailer, tbx_timer *Timer)
{
#define NLONGS 6

   char   *Subroutine_name = "transmit_timing_letter():";

   int     p;
   int     q;
   
   char   *Letter;
   int     Letter_Size;

   long    Timing[2];

   long   *Header;
   long   *Payload;

   Method  timing_max_method = {(int (*)())timing_max_fn, NULL};

#ifdef DEBUG_transmit_timing_letter
   print("transmit_timing_letter(): entry...");
#endif

   Zip_push(Mailer);

   G2_pq(p, q);

   Timing[0] = Timer -> total_sec;
   Timing[1] = Timer -> total_usec;

#ifdef DEBUG_transmit_timing_letter
   print("before combine Timing[0] = %ld, Timing[1] = %ld\n", Timing[0],
          Timing[1]);
#endif

   G2_combine(Timing, timing_max_method, 2*sizeof(int), 1);

#ifdef DEBUG_transmit_timing_letter
   print("after combine Timing[0] = %ld, Timing[1] = %ld\n", Timing[0],
          Timing[1]);
#endif

   if ((p == 0) && (q == 0))
   {
      Letter_Size = NLONGS * sizeof(long);
      Letter      = Zip_malloc(Letter_Size);
      Header      = (long *) Letter;
      Payload     = Header + 4;

      Header[0] = TIMING_LETTER;
      Header[1] = (long)Letter_Size;
      Header[2] = NLONGS;
      Header[3] = 0L;

      Payload[0] = Timing[0];
      Payload[1] = Timing[1];

      ctohl(Letter, NLONGS);
      G2_send(Letter, -1, -1);
   }

   Zip_pop();

#undef NLONGS
}

int timing_max_fn(long *ptr1, long *ptr2, int size, int nitems)
{
   if (size == 2 * sizeof(long))
   {
      if (ptr2[0] > ptr1[0])
      {
         ptr1[0] = ptr2[0];
         ptr1[1] = ptr2[1];
      }
      else if (ptr2[0] == ptr1[0])
      {
         if (ptr2[1] > ptr1[1])
            ptr1[1] = ptr2[1];
      }
   }
   else
   {
      print("timing_max_fn(): illegal size = %d vs %d", size, 2 * sizeof(long));
      exit(0);
   }
}

void receive_timing_host(ZIP_MAILER *Mailer, long *sec, long *usec)
{
   char   *Subroutine_name = "receive_timing_letter():";
   int     p;
   int     q;

   char   *Letter;

   long   *Header;
   long   *Payload;

   Zip_push(Mailer);

   G2_pq(p, q);

   Letter = (char *) G2_recvb(0, 0);

   Header = (long *)Letter;

   if (*Header != TIMING_LETTER)
   {
      print("csparsetest1a: anticipated TIMING_LETTER=%ld, received %ld",
            (long)TIMING_LETTER, *Header);

      exit(1);
   }

   Payload = Header + 4;

   *sec  = Payload[0];
   *usec = Payload[1];

   Zip_free(Letter);

   Zip_pop();
}

void transmit_col_vector_host(ZIP_MAILER *Mailer, vector_host_bin *Vector)
{
#define NLONGS 6

   char      *Subroutine_name = "transmit_col_vector_host():";
   char      *Letter;
   int        Letter_Size;

   long      *Header;

   long      *Vector_i;
   double    *Vector_Value;

   int        M = Vector -> M;
   int        m;

   int        P = Vector -> P;
   int        p;

   long       Vector_element_Count;
   long       long_Count;

   int        Index;

   element  **b = Vector -> b;
   int       *b_element_Count = Vector -> b_element_Count;
   element   *Pointer;

#ifdef DEBUG_transmit_col_vector_host
   print("%s entry...", Subroutine_name);
#endif

   Zip_push(Mailer);

   for (p = 0; p < P; p++)
   {
      Vector_element_Count = b_element_Count[p];

      long_Count = Vector_element_Count;
      long_Count += (Vector_element_Count & 0x00000001);
      Letter_Size = NLONGS * sizeof(long)
                    + long_Count * sizeof(long)
                    + Vector_element_Count * sizeof(double);

      Letter    = Zip_malloc(Letter_Size);
      Header    = (long *) Letter;
      Header[0] = VECTOR_LETTER;
      Header[1] = (long)Letter_Size;
      Header[2] = NLONGS;
      Header[3] = 0L;

      Header[4] = M;
      Header[5] = Vector_element_Count;

      htocl(Header, NLONGS);

#ifdef DEBUG_transmit_col_vector_host
      print("%s Vector_element_Count = %d\n", Subroutine_name,
            Vector_element_Count);
#endif
      if (Vector_element_Count > 0)
      {
         Vector_i = &(Header[NLONGS]);
         Vector_Value = (double *) &(Vector_i[long_Count]);

         Index = 0;
         Pointer = b[p];
         do
         {
#ifdef DEBUG_transmit_col_vector_host
            print("%s Pointer = 0x%lx\n", Subroutine_name, Pointer);
            print("%s (%d, 0, %d, %d) = %lf\n", Subroutine_name, p,
                  Pointer -> i, Pointer -> j, Pointer -> Real);
#endif
            Vector_i[Index] = Pointer -> i;
            Vector_Value[Index] = Pointer -> Real;
            Index++;
            Pointer = Pointer -> Next;
         } while (Pointer != NULL);

         if (Index < Vector_element_Count)
         {
            print("%s Lost matrix entries\n", Subroutine_name);
            print("Index = %d, Vector_element_Count = %d\n", Index,
                  Vector_element_Count);
         }
         htocl(Vector_i, Vector_element_Count);
         htocd(Vector_Value, Vector_element_Count);
      }

      G2_send(Letter, p, 0);
   }

   Zip_pop();

#undef NLONGS
}

void receive_col_vector_node(ZIP_MAILER *Mailer, vector_node_bin *Vector)
{
   char   *Subroutine_name = "receive_col_vector_node()";
   int     P;
   int     Q;
   int     p;
   int     q;

   char   *Letter;
   int     Letter_Size;

   long   *Header;
   long   *Vec_Data_Header;
   long   *Vector_i;
   double *Vector_Value;
   double *b = Vector -> b -> a -> s;
   int     b_element_count = Vector -> b -> a -> m;

   int     i;
   int     I;
   int     long_Count;

   int     serror;

#ifdef DEBUG_receive_col_vector_node
   print("%s - begin\n", Subroutine_name);
   print("%s - Vector = 0x%lx\n", Subroutine_name, Vector);
   print("%s - b = 0x%lx, m = %d\n", Subroutine_name, b, b_element_count);
#endif

   Zip_push(Mailer);

   G2_pq(p, q);
   G2_PQ(P, Q);

   if (q == 0)
   {
#ifdef DEBUG_receive_col_vector_node
      print("waiting for letter from HOST");
#endif
      Letter      = (char *) G2_recvb(-1, -1);
      Letter_Size = ((long *)Letter)[1];
   }
   else
   {
      Letter      = NULL;
      Letter_Size = 0;
   }

   /* fanout: */
   if (Q > 1)
   {
      serror = G2_row_fanout(&Letter, &Letter_Size, 0);

#ifdef DEBUG_receive_col_vector_node
   print("%s After fanout\n", Subroutine_name);
#endif
   }

#ifdef DEBUG_receive_col_vector_node
   print("%s Letter = 0x%lx Letter_Size = %d\n", Subroutine_name, Letter,
         Letter_Size);
#endif

   Header = (long *) Letter;

   if (*Header != VECTOR_LETTER)
   {
      print("%s anticipated VECTOR_LETTER=%ld, received %ld", Subroutine_name,
            (long)VECTOR_LETTER, *Header);

      exit(1);
   }

   Vec_Data_Header = Header + 4;
   Vector_i = Vec_Data_Header + 2;
   long_Count = Vec_Data_Header[1];
   long_Count += (Vec_Data_Header[1] & 0x00000001);
   Vector_Value = (double *) (Vector_i + long_Count);

#ifdef DEBUG_receive_col_vector_node
   print("%s Vec_Data_Header[1] = %d\n", Subroutine_name, Vec_Data_Header[1]);
#endif

   if (Vec_Data_Header[1] > 0)
   {
      if (Vec_Data_Header[1] != b_element_count)
      {
         print("%s Cvector length mismatch.\n", Subroutine_name);
         exit(1);
      }

      for (i = 0; i < b_element_count; i++)
      {
         I = Vector_i[i];
#ifdef DEBUG_receive_col_vector_node
         print("%s (%d, %d, %d) = %lf\n", Subroutine_name, p, q, I,
               Vector_Value[i]);
#endif
         b[I] = Vector_Value[i];
      }
   }
   else
   {
      for (i = 0; i < b_element_count; i++)
         b[i] = 0.0;
   }

   if (q == 0)
      Zip_free(Letter);
   else
      free(Letter);

   Zip_pop();
}

void transmit_row_vector_node(ZIP_MAILER *Mailer, vector_node_bin *Vector)
{
#define NLONGS 6
   char    *Subroutine_name = "transmit_row_vector_node()";
   int      P;
   int      Q;
   int      p;
   int      q;

   char    *Letter;
   int      Letter_Size;

   long    *Header;
   long    *Vector_i;
   double  *Vector_Value;
   int    (*Col_Inverse_Distribution)();
   void    *Col_Extra = Vector -> Col_Extra;
   double  *x               = Vector -> x -> a -> s;
   int      x_element_Count = Vector -> x -> a -> m;
   int      N = Vector -> x -> M;

   int      i;
   int      I;
   int      long_Count;

   int      serror;

#ifdef DEBUG_transmit_row_vector_node
   print("%s - begin\n", Subroutine_name);
   print("%s - Vector = 0x%lx\n", Subroutine_name, Vector);
   print("%s - x = 0x%lx, m = %d\n", Subroutine_name, x, x_element_Count);
#endif

   Col_Inverse_Distribution = Vector -> Col_Inverse_Distribution;

   Zip_push(Mailer);

   G2_pq(p, q);
   G2_PQ(P, Q);

   if (p == 0)
   {
      long_Count = x_element_Count;
      long_Count += (x_element_Count & 0x00000001);
      Letter_Size = NLONGS * sizeof(long) + long_Count * sizeof(long)
                    + x_element_Count * sizeof(double);

      Letter    = Zip_malloc(Letter_Size);
      Header    = (long *) Letter;
      Header[0] = VECTOR_LETTER;
      Header[1] = (long)Letter_Size;
      Header[2] = NLONGS;
      Header[3] = 0L;

      Header[4] = N;
      Header[5] = x_element_Count;

      ctohl(Header, NLONGS);

      Vector_i = &(Header[6]);
      Vector_Value = (double *) (Vector_i + long_Count);

      for (i = 0; i < x_element_Count; i++)
      {
         Vector_i[i] = (long) (*Col_Inverse_Distribution)(q, i, Q, N,
                                                          Col_Extra);
         Vector_Value[i] = x[i];
#ifdef DEBUG_transmit_row_vector_node
         print("%s (%d, %d, %d) (%d) = %lf\n", Subroutine_name, p, q, i,
               Vector_i[i], x[i]);
#endif
      }

      ctohl(Vector_i, x_element_Count);
      ctohd(Vector_Value, x_element_Count);

      G2_send(Letter, -1, -1);
   }

   Zip_pop();
#undef NLONGS
}

void receive_row_vector_host(ZIP_MAILER *Mailer, vector_host_bin *Vector)
{
   char      *Subroutine_name = "receive_row_vector_host():";
   char      *Letter;

   long      *Header;

   long      *Vector_i;
   double    *Vector_Value;

   int        Q               = Vector -> Q;
   int        q;

   long       long_Count;

   int        i;
   int        i_limit;

   element   *x               = NULL;
   int        x_element_Count = 0;
   element   *Pool            = Vector -> Pool;
   int        J;
   int        J_offset;

#ifdef DEBUG_receive_row_vector_host
   print("%s entry...", Subroutine_name);
#endif

   Zip_push(Mailer);

   for (q = 0; q < Q; q++)
   {
      Letter    = (char *) G2_recvb(0, q);
      Header    = (long *) Letter;

      long_Count = Header[5];
      long_Count += (Header[5] & 0x00000001);

#ifdef DEBUG_receive_row_vector_host
      print("%s x_local_element_count = %d\n", Subroutine_name,
            Header[5]);
#endif

      Vector_i = &(Header[6]);
      Vector_Value = (double *) (Vector_i + long_Count);

      i_limit = Header[5];
      for (i = 0; i < i_limit; i++)
      {
         J = Vector_i[i];

         J_offset = J & 0x00000001;
         J >>= 1;
         if (J_offset == 0)
         {
            add_element_real(&Pool, &x, &x_element_Count, 0, J,
                             Vector_Value[i]);
         }
         else
         {
            add_element_imag(&Pool, &x, &x_element_Count, 0, J,
                             Vector_Value[i]);
         }
#ifdef DEBUG_receive_row_vector_host
         print("%s x = 0x%lx, x_element_Count = %d\n", Subroutine_name,
               x, x_element_Count);
         print("%s (%d, %d, %d) (%d) = %lf\n", Subroutine_name, 0, q, i,
               Vector_i[i], Vector_Value[i]);
#endif
      }
   }

   Vector -> Col_major_x = x;
   Vector -> x_element_Count = x_element_Count;
   Vector -> Pool = Pool;

#ifdef DEBUG_receive_row_vector_host
   print("%s x_element_Count = %d\n", Subroutine_name, x_element_Count);
#endif

   Zip_pop();
}

